/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_CONV_H_
#define AIDGE_CORE_OPERATOR_CONV_H_

#include <array>
#include <cmath>
#include <numeric>
#include <vector>

#include "aidge/data/Tensor.hpp"
#include "aidge/graph/Node.hpp"
#include "aidge/operator/Operator.hpp"
#include "aidge/operator/Producer.hpp"
#include "aidge/utils/Parameter.hpp"
#include "aidge/utils/Registrar.hpp"
#include "aidge/utils/Types.h"

namespace Aidge {
enum class ConvParam { StrideDims, DilationDims, InChannels, OutChannels, KernelDims, PaddingDims };

template <DimIdx_t DIM>
class Conv_Op : public Operator,
                public Registrable<Conv_Op<DIM>, std::string, std::unique_ptr<OperatorImpl>(const Conv_Op<DIM> &)>,
                public Parameterizable<ConvParam, std::array<DimSize_t, DIM>, std::array<DimSize_t, DIM>, DimSize_t,
                                       DimSize_t, std::array<DimSize_t, DIM>, std::array<DimSize_t, (DIM<<1) >> {
public:
    // FIXME: change accessibility
    std::array<std::shared_ptr<Tensor>, 3> mInputs = {std::make_shared<Tensor>(), std::make_shared<Tensor>(),
                                                      std::make_shared<Tensor>()};
    const std::shared_ptr<Tensor> mOutput = std::make_shared<Tensor>();

   public:
    static constexpr const char *Type = "Conv";

    Conv_Op() = delete;

    using Parameterizable_ = Parameterizable<ConvParam, std::array<DimSize_t, DIM>, std::array<DimSize_t, DIM>,
                                             DimSize_t, DimSize_t, std::array<DimSize_t, DIM>, std::array<DimSize_t, (DIM<<1) >>;
    template <ConvParam e>
    using param = typename Parameterizable_::template param<e>;

    constexpr Conv_Op(DimSize_t in_channels,
                      DimSize_t out_channels,
                      const std::array<DimSize_t, DIM> &kernel_dims,
                      const std::array<DimSize_t, DIM> &stride_dims = create_array<DimSize_t,DIM>(1),
                      const std::array<DimSize_t, (DIM<<1)> &padding_dims = create_array<DimSize_t,(DIM<<1)>(0),
                      const std::array<DimSize_t, DIM> &dilation_dims = create_array<DimSize_t,DIM>(1))
        : Operator(Type),
          Parameterizable_(param<ConvParam::StrideDims>(stride_dims),
                           param<ConvParam::DilationDims>(dilation_dims),
                           param<ConvParam::InChannels>(in_channels),
                           param<ConvParam::OutChannels>(out_channels),
                           param<ConvParam::KernelDims>(kernel_dims),
                           param<ConvParam::PaddingDims>(padding_dims)),
          mOutput(std::make_shared<Tensor>()) {
        setDatatype(DataType::Float32);
    }

    // Data operator[](const char* inputName) override final {
    //     std::shared_ptr<Tensor> in = (strcmp(inputName, "data")) ? mInputs[0] :
    //         (strcmp(inputName, "weight") ? mInputs[1] :
    //         (strcmp(inputName, "bias") ? mInputs[2] :
    //         nullptr));
    //     assert((in!=nullptr) && "No such parameter");
    //     return *in;
    // }

    // std::shared_ptr<Conv_Op> clone() const override final {

    // }

    constexpr void associateInput(const IOIndex_t inputIdx, std::shared_ptr<Data> data) override final {
        assert(inputIdx < 3 && "operators supports only 3 inputs");
        assert(strcmp(data->type(), Tensor::Type) == 0 && "input data must be of Tensor type");

        mInputs[inputIdx] = std::dynamic_pointer_cast<Tensor>(data);
    }

    constexpr void computeOutputDims() override final {
        if (!mInputs[0]->empty()) {
            std::array<DimSize_t, DIM + 2> outputDims = {};

            for (std::size_t dim = 0; dim < this->template get<ConvParam::KernelDims>().size() ; ++dim) {
                const DimSize_t kernelExtent = this->template get<ConvParam::DilationDims>()[dim] *
                                                       (this->template get<ConvParam::KernelDims>()[dim] - 1) +
                                               1;

                outputDims[dim+2] = 1 + static_cast<DimSize_t>(
                        floor(static_cast<float>(mInputs[0]->dims()[dim+2] - kernelExtent +
                                                 this->template get<ConvParam::PaddingDims>()[dim] +
                                                 this->template get<ConvParam::PaddingDims>()[dim+DIM]) /
                              static_cast<float>(this->template get<ConvParam::StrideDims>()[dim])));
            }

            outputDims[1] = this->template get<ConvParam::OutChannels>();
            outputDims[0] = mInputs[0]->dims()[0];
            mOutput->resize(outputDims);
        }
    }

    bool outputDimsForwarded() const override final { return !(mOutput->empty()); }


    inline Tensor& input(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < 3 && "operators supports only 3 inputs");
        return *(mInputs[inputIdx].get()); }
    inline Tensor& output(const IOIndex_t /*outputIdx*/) const override final { return *(mOutput.get()); }


    inline std::shared_ptr<Tensor> getInput(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < 3 && "Conv Operators supports only 3 inputs");
        return mInputs[inputIdx];
    }
    inline std::shared_ptr<Tensor> getOutput(const IOIndex_t outputIdx) const override final {
        assert((outputIdx == 0) && "Conv Operator has only 1 output");
        (void) outputIdx; // avoid unused warning
        return mOutput;
    }


    std::shared_ptr<Data> getRawInput(const IOIndex_t inputIdx) const override final {
        assert(inputIdx < 3 && "operators supports only 3 inputs");
        return std::static_pointer_cast<Data>(mInputs[inputIdx]);
    }
    std::shared_ptr<Data> getRawOutput(const IOIndex_t outputIdx) const override final {
        assert(outputIdx == 0 && "operator supports only 1 output");
        (void) outputIdx; // avoid unused warning
        return std::static_pointer_cast<Data>(mOutput);
    }


    void setBackend(const std::string &name) {
        mImpl = Registrar<Conv_Op<DIM>>::create(name)(*this);
        mOutput->setBackend(name);

        // FIXME: temporary workaround
        mInputs[1]->setBackend(name);
        mInputs[2]->setBackend(name);
    }

    void setDatatype(const DataType &datatype) {
        mOutput->setDatatype(datatype);

        // FIXME: temporary workaround
        mInputs[0]->setDatatype(datatype);
        mInputs[1]->setDatatype(datatype);
        mInputs[2]->setDatatype(datatype);
    }

    inline IOIndex_t nbInputs() const noexcept override final { return 3; }
    inline IOIndex_t nbDataInputs() const noexcept override final { return 1; }
    inline IOIndex_t nbOutputs() const noexcept override final { return 1; }
};

template <std::array<DimSize_t, 1>::size_type DIM>
inline std::shared_ptr<Node> Conv(DimSize_t in_channels,
                                  DimSize_t out_channels,
                                  const std::array<DimSize_t, DIM> &kernel_dims,
                                  const std::string& name = "",
                                  const std::array<DimSize_t, DIM> &stride_dims = create_array<DimSize_t,DIM>(1),
                                  const std::array<DimSize_t, (DIM<<1)> &padding_dims = create_array<DimSize_t,(DIM<<1)>(0),
                                  const std::array<DimSize_t, DIM> &dilation_dims = create_array<DimSize_t,DIM>(1)) {
    // FIXME: properly handle default w&b initialization in every cases
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Conv, not supported");
    auto conv = std::make_shared<Node>(std::make_shared<Conv_Op<static_cast<DimIdx_t>(DIM)>>(in_channels, out_channels, kernel_dims, stride_dims, padding_dims, dilation_dims), name);
    // addProducer(conv, 1, append(append(kernel_dims, in_channels), out_channels), "w");
    addProducer(conv, 1, append(out_channels, append(in_channels, kernel_dims)), "w");
    addProducer(conv, 2, {out_channels}, "b");
    return conv;
}

template <DimSize_t DIM>
inline std::shared_ptr<Node> Conv(
    DimSize_t in_channels,
    DimSize_t out_channels,
    DimSize_t const (&kernel_dims)[DIM],
    const std::string& name = "",
    const std::array<DimSize_t, DIM> &stride_dims = create_array<DimSize_t,DIM>(1),
    const std::array<DimSize_t, (DIM<<1)> &padding_dims = create_array<DimSize_t,(DIM<<1)>(0),
    const std::array<DimSize_t, DIM> &dilation_dims = create_array<DimSize_t,DIM>(1)) {
    static_assert(DIM<=MaxDim,"Too many kernel dimensions required by Conv, not supported");
    return Conv(in_channels, out_channels, to_array(kernel_dims), name, stride_dims, padding_dims, dilation_dims);
}
}  // namespace Aidge

namespace {
template <>
const char *const EnumStrings<Aidge::ConvParam>::data[] = {
    "StrideDims",
    "DilationDims",
    "InChannels",
    "OutChannels",
    "KernelDims",
    "PaddingDims"
};
}

#endif /* AIDGE_CORE_OPERATOR_CONV_H_ */
