/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CORE_OPERATOR_OPERATOR_H_
#define AIDGE_CORE_OPERATOR_OPERATOR_H_

#include <memory>
#include <string>
#include <vector>

#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/data/Data.hpp"
#include "aidge/data/Tensor.hpp"
#include "aidge/utils/Types.h"
#include "aidge/hook/hook.hpp"

namespace Aidge {

class Operator : public std::enable_shared_from_this<Operator> {
protected:
  std::unique_ptr<OperatorImpl> mImpl; // implementation of the operator
  std::map<std::string, std::shared_ptr<Hook>> mHooks;

private:
  std::string mType;

public:
  Operator() = delete;
  Operator(const char* type) : mType(type) {}
  virtual ~Operator();


public:

    virtual void associateInput(const IOIndex_t inputIdx, std::shared_ptr<Data> data) = 0;
    virtual void computeOutputDims() = 0;
    virtual bool outputDimsForwarded() const = 0;
    virtual std::shared_ptr<Data> getRawInput(const IOIndex_t inputIdx) const = 0;
    virtual std::shared_ptr<Tensor> getInput(const IOIndex_t inputIdx) const = 0;
    virtual Tensor& input(const IOIndex_t /*inputIdx*/) const = 0;
    virtual std::shared_ptr<Data> getRawOutput(const IOIndex_t outputIdx) const = 0;
    virtual std::shared_ptr<Tensor> getOutput(const IOIndex_t outputIdx) const = 0;
    virtual Tensor& output(const IOIndex_t /*outputIdx*/) const = 0;

    std::shared_ptr<Hook> getHook(std::string hookName) {
        return mHooks[hookName];
    }
    void addHook(std::string hookName) {
        mHooks.insert(std::pair<std::string, std::shared_ptr<Hook>>(hookName,Registrar<Hook>::create({hookName})(shared_from_this())));
    }

    void runHooks() const;

///////////////////////////////////////////////////////
//        IMPLEMENTATION
///////////////////////////////////////////////////////

    virtual void setBackend(const std::string& name) = 0;
    virtual void setDatatype(const DataType& datatype) = 0;

    /**
     * @brief Minimum amount of data from a specific input for one computation pass.
     * @param inputIdx Index of the input analysed.
     * @return NbElts_t
     */
    NbElts_t getNbRequiredData(const IOIndex_t inputIdx) const;

    /**
     * @brief Amount of data from a specific input actually used in one computation pass.
     *
     * @param inputIdx Index of the input analysed.
     * @return NbElts_t
     */
    NbElts_t getNbConsumedData(const IOIndex_t inputIdx) const;

    /**
     * @brief Amount of data ready to be used on a specific output.
     *
     * @param outputIdx Index of the output analysed.
     * @return NbElts_t
     */
    NbElts_t getNbProducedData(const IOIndex_t outputIdx) const;

    void updateConsummerProducer();

    virtual void forward();

    virtual void backward();

///////////////////////////////////////////////////////
//        INNER
///////////////////////////////////////////////////////

    std::string type() const {
        return mType;
    }

    virtual IOIndex_t nbInputs() const noexcept = 0;
    virtual IOIndex_t nbDataInputs() const noexcept = 0;
    virtual IOIndex_t nbOutputs() const noexcept = 0;
};
} // namespace Aidge

#endif /* AIDGE_CORE_OPERATOR_OPERATOR_H_ */
