
/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_ANY_H_
#define AIDGE_ANY_H_


#include <typeinfo>    // typeid
#include <type_traits> // std::enable_if_t, std::decay_t, std::is_same, std::is_copy_constructible, std::remove_cv, std::remove_reference
#include <assert.h>
#include <new>

class _any {
private:
    /// @brief Operation to perform on the object.
    enum _Op { _Op_access, _Op_get_type_info, _Op_clone, _Op_destroy };

    union _Arg {
        const std::type_info* _M_typeinfo;
        _any* _M_any;
    };

    /// @brief Stored data without type information.
    void* _M_data;

    /// @brief Member function to perform type-related computations on stored data.
    void (*_M_manager)(_Op, const _any*, _Arg*);

public:
    /// @brief Class to centralize functions and type information in a memory efficient way.
    /// @tparam Tp Decayed stored type.
    template <typename Tp>
    struct Manager {
        static void manage(_Op which, const _any* __any, _Arg* __arg) {
            auto ptr = static_cast<const Tp*>(__any->_M_data);
            switch (which)
            {
            case _Op_get_type_info:
                __arg->_M_typeinfo = &typeid(Tp);
                break;
            case _Op_clone:
                __arg->_M_any->_M_data = new Tp(*ptr);
                __arg->_M_any->_M_manager = __any->_M_manager;
                break;
            case _Op_destroy:
                delete ptr;
                break;
            }

        }
        static Tp* access(const _any* __any) {
            return static_cast<Tp*>(__any->_M_data);
        }

        // template <typename Up>
        // static void create(void* data, Up&& value) {
        //     data = new Tp(std::forward<Up>(value));
        // }
    };

private:
    template<typename _Tp, typename _VTp = std::decay_t<_Tp>>
    using _Decay_if_not_any = std::enable_if_t<!std::is_same<_VTp, _any>::value, _VTp>;

public:
    /// @brief Default constructor
    _any() noexcept : _M_manager(nullptr) { }

    /// @brief Copy constructor
    /// @param __other
    _any(const _any& __other)
    {
        if (!__other._M_manager)
            _M_manager = nullptr;
        else
        {
            _Arg __arg;
            __arg._M_any = this;
            __other._M_manager(_Op_clone, &__other, &__arg);
        }
    }

    /// @brief Move constructor
    /// @param __other
    _any(_any&& __other)
    {
        if (!__other._M_manager)
            _M_manager = nullptr;
        else
        {
            _M_data = __other._M_data;
            _M_manager = __other._M_manager;
            const_cast<_any*>(&__other)->_M_manager = nullptr;
        }
    }

    /// @brief By-value constructor.
    /// @tparam T Data type.
    /// @tparam VT Decayed data type.
    /// @param value
    template<typename T, typename VT = _Decay_if_not_any<T>, std::enable_if_t<std::is_copy_constructible<VT>::value, bool> = true>
    explicit _any(T&& value)
        : _M_manager(&Manager<VT>::manage),
          _M_data(new VT{std::forward<T>(value)})
    {}

    ~_any()
    {
        if(_M_manager) {
            _M_manager(_Op_destroy, this, nullptr);
            _M_manager = nullptr;
        }
    }

    /// @brief Access type id of the value currently stored
    /// @return
    const std::type_info& type() const
    {
        if (!_M_manager)
            return typeid(void);
        _Arg __arg;
        _M_manager(_Op_get_type_info, this, &__arg);
        return *__arg._M_typeinfo;
    }
};

/// @brief Access value stored in the object converted in the template type if possible.
/// @tparam _ValueType
/// @param __any
/// @return Stored value.
template<typename _ValueType>
inline _ValueType any_cast(const _any& __any)
{
    using _Up =  std::remove_cv_t<std::remove_reference_t<_ValueType>>;
    assert((std::__or_<std::is_reference<_ValueType>, std::is_copy_constructible<_ValueType>>::value && "Template argument must be a reference or CopyConstructible type"));
    assert((std::is_constructible<_ValueType, const _Up&>::value && "Template argument must be constructible from a const value."));
    assert(std::is_object<_Up>::value);
    assert(__any.type() == typeid(_Up));
    auto __p = static_cast<_Up*>(__any._M_data);
    if (__p)
        return static_cast<_ValueType>(*__p);
    throw std::bad_cast();
}

#endif /* AIDGE_ANY_H_ */