/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CPARAMETER_H_
#define AIDGE_CPARAMETER_H_

#include <map>
#include <vector>
#include <type_traits>
#include <typeinfo>
#include <assert.h>

#include "aidge/utils/Any.hpp"


namespace Aidge {

///\todo store also a fix-sized code that indicates the type
///\todo managing complex types or excluding non-trivial, non-aggregate types
class CParameter {
private:
    template<typename _ValueType>
    inline _ValueType& any_cast_ref(const _any& __any)
    {
        using _Up =  std::remove_cv_t<std::remove_reference_t<_ValueType>>;
        assert(((std::is_reference<_ValueType>::value || std::is_copy_constructible<_ValueType>::value) && "Template argument must be a reference or CopyConstructible type"));
        assert((std::is_constructible<_ValueType, const _Up&>::value && "Template argument must be constructible from a const value."));
        assert(std::is_object<_Up>::value);
        assert(__any.type() == typeid(_Up));
        if (_any::Manager<_Up>::access(&__any)) { // assess if _any object is empty
            return *static_cast<_ValueType*>(_any::Manager<_Up>::access(&__any));
        }
        throw std::bad_cast();
    }
public:
    // not copyable, not movable
    CParameter(CParameter const &) = delete;
    CParameter(CParameter &&) = delete;
    CParameter &operator=(CParameter const &) = delete;
    CParameter &operator=(CParameter &&) = delete;
    CParameter() : m_Params({}){};
    ~CParameter() = default;

    /**
     * \brief Returning a parameter identified by its name
     * \tparam T expected parameter type
     * \param i_ParamName Parameter name
     * \details assert if T is not the actual parameter type, if the parameter does not
     *  exist or interna parameter position is invalid.
     * \todo Returning a T const& ? But dangerous => the client may get an address within
     *  param buffer that will get invalid after the CParam death.
     * \note at() throws if the parameter does not exist, using find to test for parameter existance
     */
    template<class T> T& Get(const std::string i_ParamName)
    {
        return any_cast_ref<T>(m_Buffer[m_Params.at(i_ParamName)]);
    }

    // template<class T> const T& Get(const std::string i_ParamName) const
    // {
    //     return any_cast<T>(m_Buffer[m_Params.at(i_ParamName)]);
    // }

    ///\brief Add a parameter value, identified by its name
    ///\tparam T expected parameter type
    ///\param i_ParamName Parameter name
    ///\param i_Value Parameter value
    ///\todo Pass i_Value by ref if large or not trivial
    ///\bug If parameter already exists, its value is changed but written in the
    /// internal buffer in a new location (previous value is still in memory at its previous location)
    template<class T> void Add(const std::string &i_ParamName, T&& i_Value)
    {
        m_Params[i_ParamName] = m_Buffer.size(); // Copy pointer offset
        m_Buffer.push_back(_any(std::forward<T>(i_Value)));
    }


    std::string getParamType(std::string const &i_ParamName){
        return m_Buffer[m_Params.at(i_ParamName)].type().name();
    }

    std::vector<std::string> getParametersName(){
        std::vector<std::string> parametersName;
        for(auto const& it: m_Params)
            parametersName.push_back(it.first);
        return parametersName;
    }

private:
    std::map<std::string, std::size_t> m_Params; // { Param name : offset }

    ///\brief All raw pointers to parameters values concatenated. Use custom any class compatible with C++14.
    std::vector<_any> m_Buffer = {};
};

}

#endif /* AIDGE_CPARAMETER_H_ */
