/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/
#ifdef PYBIND
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

#include <string>
#include <vector>
#include <array>

#include "aidge/utils/Parameter.hpp"
#include "aidge/backend/OperatorImpl.hpp"
#include "aidge/operator/AvgPooling.hpp"
#include "aidge/operator/Operator.hpp"
#include "aidge/utils/Types.h"
#include "aidge/data/Tensor.hpp"

namespace py = pybind11;
namespace Aidge {

template <DimIdx_t DIM> void declare_AvgPoolingOp(py::module &m) {
  py::class_<AvgPooling_Op<DIM>, std::shared_ptr<AvgPooling_Op<DIM>>, Operator, PyAbstractParametrizable>(
    m, ("AvgPoolingOp" + std::to_string(DIM) + "D").c_str(),
    py::multiple_inheritance())
  .def(py::init<const std::array<DimSize_t, DIM> &,
                const std::array<DimSize_t, DIM> &,
                const std::array<DimSize_t, (DIM<<1)> &>(),
        py::arg("kernel_dims"),
        py::arg("stride_dims"),
        py::arg("padding_dims"));
  
  m.def(("AvgPooling" + std::to_string(DIM) + "D").c_str(), [](const std::vector<DimSize_t>& kernel_dims, 
                                                                  const std::string& name,
                                                                  const std::vector<DimSize_t> &stride_dims,
                                                                  const std::vector<DimSize_t> &padding_dims) {
        // Lambda function wrapper because PyBind fails to convert const array.
        // So we use a vector that we convert in this function to a const DimeSize_t [DIM] array. 
        if (kernel_dims.size() != DIM) {
            throw std::runtime_error("kernel_dims size [" + std::to_string(kernel_dims.size()) + "] does not match DIM [" + std::to_string(DIM) +"]");
        }
        if (stride_dims.size() != DIM) {
            throw std::runtime_error("stride_dims size [" + std::to_string(stride_dims.size()) + "] does not match DIM [" + std::to_string(DIM) +"]");
        }
        if (padding_dims.size() != (DIM<<1)) {
            throw std::runtime_error("padding_dims size [" + std::to_string(padding_dims.size()) + "] does not match DIM [" + std::to_string(DIM<<1) +"]");
        }
        DimSize_t tmp_kernel_dims_array[DIM];
        for (size_t i = 0; i < DIM; ++i) {
            tmp_kernel_dims_array[i] = kernel_dims[i];
        }
        DimSize_t tmp_stride_dims_array[DIM];
        for (size_t i = 0; i < DIM; ++i) {
            tmp_stride_dims_array[i] = stride_dims[i];
        }
        DimSize_t tmp_padding_dims_array[DIM<<1];
        for (size_t i = 0; i < (DIM<<1); ++i) {
            tmp_padding_dims_array[i] = padding_dims[i];
        }
        const DimSize_t (&kernel_dims_array)[DIM] = tmp_kernel_dims_array;
        const DimSize_t (&stride_dims_array)[DIM] = tmp_stride_dims_array;
        const DimSize_t (&padding_dims_array)[DIM<<1] = tmp_padding_dims_array;
        return AvgPooling<DIM>(to_array(kernel_dims_array), name, to_array(stride_dims_array), to_array(padding_dims_array));
    }, py::arg("kernel_dims"),
       py::arg("name") = "",
       py::arg("stride_dims") = std::vector<DimSize_t>(DIM,1),
       py::arg("padding_dims") = std::vector<DimSize_t>(DIM<<1,0));
  
}


void init_AvgPooling(py::module &m) {
  declare_AvgPoolingOp<1>(m);
  declare_AvgPoolingOp<2>(m);
  declare_AvgPoolingOp<3>(m);
 
  // FIXME:
  // m.def("AvgPooling1D", static_cast<NodeAPI(*)(const char*, int, int, int const
  // (&)[1])>(&AvgPooling));
}
} // namespace Aidge
#endif